#!/usr/bin/perl -Tw
# Written by Dave Ackley in 2011-2012 and placed in the public domain.

# Gah!  Perl has messed with its sorting algorithms in COMPLETELY
# NON-BACKWARD-COMPATIBLE WAYS!  So require our version (more or less,
# anyway) to try to stabilize these results!
require 5.10.0;  
use List::Util 'shuffle';

{
   use sort qw(defaults _quicksort); # force quicksort
   no sort "stable";                 # stability not wanted
   BEGIN { 
       sort::current eq "quicksort" 
           or die "Expected 'quicksort', got '".sort::current."'";
   }
   sub useQsort {
       return sort dcmp @_;
   }
}

{
   use sort qw(defaults _mergesort); # force mergesort
   BEGIN { 
       sort::current eq "mergesort" 
           or die "Expected 'mergesort', got '".sort::current."'";
   }
   sub useMergesort {
       return sort dcmp @_;
   }
}

# output: space-separated data for gnuplot (via makeGraph.pl)
# Columns:
#  0: tries: times to feed back and re-sort output before scoring it
#  1: compFailsPct: i.i.d. percentage of comparisons that fail
#  2: avgq: average quicksort score over $samplesToAvg runs
#  3: devq: sample standard deviation of quicksort score
#  4: avgb: average bubblesort score over $samplesToAvg runs
#  5: devb: sample standard deviation of bubblesort score
#  6: avgm: average perl mergesort score over $samplesToAvg runs
#  7: devm: sample standard deviation of perl mergesort score
#  8: avgmm: average my mergesort score over $samplesToAvg runs
#  9: devmm: sample standard deviation of my mergesort score
# 10: avgn: average no sort score over $samplesToAvg runs
# 11: devn: sample standard deviation of no sort score
# 12: cmpq: total comparisons over $samplesToAvg runs of quicksort
# 13: cmpb: total comparisons over $samplesToAvg runs of bubblesort
# 14: cmpm: total comparisons over $samplesToAvg runs of perl mergesort
# 15: cmpmm: total comparisons over $samplesToAvg runs of my mergesort
# 16: cmpn: total comparisons over $samplesToAvg runs of no sort

$|=1;

if (scalar(@ARGV)==0) {
    srand(1);              # standardize result by default
} elsif ($ARGV[0] =~ /\D/) {
    die "Non-int arg supplied '$ARGV[0]'";
} elsif ($ARGV[0] > 0) {   # use positive arg as seed
    srand($ARGV[0]);
} else {                   # 0 arg is flag to randomize..
    my $seed = time();
    srand($seed);
    print STDERR "$0 $seed\n";  # ..but provide info to rerun.
}

my $dcmpCalls = 0;
my $lastRawDcmp = 0;
my $max = 52;
my $samplesToAvg = 1000;

my $compFailsPct;   # Percentage chance of failed (random) comparison result
my $compStuckPct;   # Percentage chance of failed (stuck) comparison result
my $asks;           # How many times to (raw) compare immediately before deciding
my $tries;          # How many times to resort the alg. output before scoring

# Do a range of random fail rates (this takes a while):
# my @failPcts = (0, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1.0, 2.0, 5.0, 10.0, 20.0);
my @failPcts = (0, 10.0);

# Do a range of stuck rates (this takes a while):
#my @stuckPcts = (0, 10.0, 20.0, 30.0, 40.0, 50.0);
my @stuckPcts = (0, 50.0);

# Do a range of asks
#my @askCounts = (1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12);
my @askCounts = (1, 6);

# Loop back and resort output up to this many times (this takes a while):
# my $maxTries = 20;
my $maxTries = 1;

for my $i (@failPcts) {
    $compFailsPct = $i;
    for my $j (@stuckPcts) {
        $compStuckPct = $j;
        for my $k (@askCounts) {
            $asks = $k;
            for ($tries = 1; $tries <= $maxTries; ++$tries) {
                genData();
            }
        }
    }
}

sub genData {
    my ($avgq,$devq,$cmpq) = genDataBy(\&useQsort);
    my ($avgb,$devb,$cmpb) = genDataBy(\&useBsort);
    my ($avgm,$devm,$cmpm) = genDataBy(\&useMergesort);
    my ($avgmm,$devmm,$cmpmm) = genDataBy(\&useMyMergesort);
    my ($avgn,$devn,$cmpn) = genDataBy(\&useNoSort);
    #      0      1     2             3              4     5     6     7      8      9     10    11    12     13     14    15    16    17     18
    print "$tries $asks $compFailsPct $compStuckPct  $avgq $avgm $avgb $avgmm $avgn  $devq $devm $devb $devmm $devn  $cmpq $cmpm $cmpb $cmpmm $cmpn\n";
}

sub genDataBy {
    my $func = shift;
    my $sum = 0.0;
    my $sumsq = 0.0;
    $dcmpCalls = 0;
    $lastRawDcmp = 0;
    for (my $s = 0; $s < $samplesToAvg; ++$s) {
        my $err = genSample($func);
        $sum += $err;
        $sumsq += $err*$err;
    }
    my $mean = $sum/$samplesToAvg;
    my $dev = sqrt($sumsq/($samplesToAvg-1)-$mean*$mean);
    return ($mean,$dev,$dcmpCalls);
}

sub genSample {
    my $func = shift;
    my @n = shuffle(1..$max);
    for (my $t = 0; $t < $tries; ++$t) {
        @n = &{$func}(@n);
    }
    return scoreOutput($max,@n);
}

sub useMyMergesort {
    return mergesort(@_);
}

sub useNoSort {
    return @_;
}

sub useBsort {
    return bsort(@_);
}
sub withPct {
    return withOdds(shift,100);
}
sub withOdds {
    my ($chance,$outOf) = @_;
    return int(rand()*$outOf) < $chance;
}

sub dcmpRaw {
    ++$dcmpCalls;
    return $lastRawDcmp
        if withPct($compStuckPct);
    $lastRawDcmp = ($b <=> $a);
    $lastRawDcmp = int(rand()*3)-1  # -1,0,1 equally likely
        if withPct($compFailsPct);
    return $lastRawDcmp;
}

sub dcmp {
    my ($neg,$eql,$gtr) = (0,0,0);

    for (my $i = 0; $i < $asks; ++$i) {
        my $result = dcmpRaw();
        if ($result < 0) {
            ++$neg;
        } elsif ($result == 0) {
            ++$eql;
        } else {
            ++$gtr;
        }
    }

    my $result = -1;
    my $count = $neg;
    my $outOf = 2;
    if ($eql > $count) {
        $result = 0;
        $count = $eql;
    } elsif ($eql == $count) {
        if (withOdds(1,$outOf)) {
            $result = 0;
        }
        $outOf = 3;
    }
    if ($gtr > $count || $gtr == $count && withOdds(1,$outOf)) {
        $result = 1;
    }
    return $result;
}

sub bsort {
    my (@data) = @_;
    my $n = scalar(@data);
    for (1 .. $n) {
        for (my $j = 0; $j < $n-1; ++$j) {
            local($a) = $data[$j];
            local($b) = $data[$j+1];
            if (dcmpRaw() > 0) {     # NOTE bsort uses dcmpRaw always!  No armoring!
                my $t = $data[$j+1];
                $data[$j+1] = $data[$j];
                $data[$j] = $t;
            }
        }
    }
    return @data;
}

# I wrote this before I knew perl had switched to merge by default! 
sub mergesort {
    my (@data) = @_;
    my $n = scalar(@data);
    return @data if ($n < 2);
    my @list2 = splice(@data,$n/2);
    my @list1 = @data;
    @list1 = mergesort(@list1);
    @list2 = mergesort(@list2);

    my @ret;
    while (scalar(@list1) || scalar(@list2)) {
        my $take1 = 1;
        if (!scalar(@list1)) {
            $take1 = 0;
        } elsif (scalar(@list2)) {
            local($a) = $list1[0];
            local($b) = $list2[0];
            if (dcmp() > 0) {
                $take1 = 0;
            }
        }
        if ($take1) {
            push @ret, shift @list1;
        } else {
            push @ret, shift @list2;
        }
    }
    return @ret;
}

sub scoreOutput {
    my $max = shift;
    my @output = @_;
    my $error = 0;
    for (my $i = $max; $i > 0; --$i) {
        my $next = shift @output;
        my $err = $i-$next;
        $error += abs($err);
    }
    return $error;
}

